// local includes
#include "OpenGLColorgradientWidget.h"
#include "Coordinaten.h"

#include <QColor>
#include "GLM/gtx/type_ptr.hpp"
#include <glm/gtx/transform2.hpp>
#include <glm/gtc/matrix_projection.hpp>

using std::cout;
using std::endl;
using std::cerr;

OpenGLColorgradientWidget::OpenGLColorgradientWidget(QWidget* parent) : QGLWidget(parent), textureLoaded(false), textureHandle(0), textureParticleHandle(0), m_mode(0)
{
	this->vertices = new glm::vec4[4];
}

OpenGLColorgradientWidget::~OpenGLColorgradientWidget()
{
    delete shader;
	delete[] vertices;
}

void OpenGLColorgradientWidget::initializeGL()
{
	checkGLError();
	std::cout << "initializeGL start" << std::endl;

	// --- Do some setup for OpenGL ---
	// enable back-face culling
	glEnable(GL_CULL_FACE);
	glEnable(GL_BLEND);
	glDepthMask(GL_TRUE);
	glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
	glCullFace(GL_BACK);
	glShadeModel(GL_SMOOTH);

	// enable depth-buffer
    glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);

	//set clear color
	qglClearColor(Qt::white);

	//set Viewport is done in resizeGL

	// --- init Shader ---
	this->shader = new Shader();
	this->shader->loadCompileAndLinkShader("shader/colorgradientShader.vert", "shader/colorgradientShader.frag");

	// --- done with initializing stuff ---
	checkGLError();
	std::cout << "initializeGL end" << std::endl;
}

void OpenGLColorgradientWidget::resizeGL(int width, int height)
{
	checkGLError();
	std::cout << "resizeGL start" << std::endl;

	glViewport(0, 0, width, height);

	this->projectionMatrix = glm::ortho((float)-(width/2), (float)(width/2), (float)-(height/2), (float)(height/2), -100.0f, 100.0f);

	this->screenWidth = width;
	this->screenHeight = height;

	this->calculateVertexCoord();

	updateGL();

	checkGLError();
	std::cout << "resizeGL end" << std::endl;
}

void OpenGLColorgradientWidget::paintGL()
{
	this->checkGLError();
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	if (this->textureLoaded) 
	{
		// Bind shader
		glUseProgram(this->shader->getShaderProgram());

		//bind texture
		glActiveTexture(GL_TEXTURE0); 
		if(m_mode == 0)	glBindTexture(GL_TEXTURE_2D, textureHandle);
		else			glBindTexture(GL_TEXTURE_2D, textureParticleHandle);
		GLuint texLocation = glGetUniformLocation(this->shader->getShaderProgram(), "colorTexture");
		glUniform1i(texLocation, 0);

		// set matrix-uniforms
		GLint persp_uniform = glGetUniformLocation(this->shader->getShaderProgram(), "perspective");
		glUniformMatrix4fv(persp_uniform, 1, GL_FALSE, glm::gtx::type_ptr::value_ptr(this->projectionMatrix));

		// set attributes
		GLint vertex_attrib = glGetAttribLocation(this->shader->getShaderProgram(), "vertex");
		GLint texCoord_attrib  = glGetAttribLocation(this->shader->getShaderProgram(), "texCoord");

		glEnableVertexAttribArray(vertex_attrib);
		glEnableVertexAttribArray(texCoord_attrib);

		glVertexAttribPointer(vertex_attrib, 4, GL_FLOAT, GL_FALSE, 0, this->vertices);
		glVertexAttribPointer(texCoord_attrib,  2, GL_FLOAT, GL_FALSE, 0, quadTexCoord);

		// draw the geometry
		glDrawElements(GL_TRIANGLES, 6, GL_UNSIGNED_INT, quadIndices);

		// disable attribute-pointers
		glDisableVertexAttribArray(vertex_attrib);
		glDisableVertexAttribArray(texCoord_attrib);

		// unbind shader-program
		glUseProgram(0);

		glActiveTexture(GL_TEXTURE0);
		glBindTexture(GL_TEXTURE_2D, 0);
	}

	this->checkGLError();
}

void OpenGLColorgradientWidget::generateTransferTexture(float* rgbaValues, int textureSize, int type)
{
	checkGLError();
	makeCurrent();

	m_mode = type;

	glActiveTexture(GL_TEXTURE0); 
	if(textureHandle == 0 && m_mode == 0)
	{		
		glGenTextures(1,&textureHandle);
		glBindTexture(GL_TEXTURE_2D, textureHandle);

		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, textureSize, 1, 0, GL_RGBA, GL_FLOAT, rgbaValues);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	}
	else if(textureParticleHandle == 0 && m_mode == 1)
	{
		glGenTextures(1,&textureParticleHandle);
		glBindTexture(GL_TEXTURE_2D, textureParticleHandle);
		
		glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA32F, textureSize, 1, 0, GL_RGBA, GL_FLOAT, rgbaValues);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MIN_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D,GL_TEXTURE_MAG_FILTER,GL_LINEAR);
		glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
	}
	else
	{
		//So ist es korrekt
		glActiveTexture(GL_TEXTURE0); 
		if(m_mode == 0)	glBindTexture(GL_TEXTURE_2D, textureHandle);
		else			glBindTexture(GL_TEXTURE_2D, textureParticleHandle);
		glTexSubImage2D(GL_TEXTURE_2D, 0, 0, 0, textureSize, 1, GL_RGBA, GL_FLOAT, rgbaValues);
	}

	this->textureLoaded = true;
	this->calculateVertexCoord();
	updateGL();

	checkGLError();
}

void OpenGLColorgradientWidget::resetTransferTexture()
{
	makeCurrent();
	glDeleteTextures(1,&textureHandle);
	glDeleteTextures(1,&textureParticleHandle);
	this->textureLoaded = false;
	updateGL();
}

void OpenGLColorgradientWidget::calculateVertexCoord()
{
	//calculate and set texture position
	float posX = -this->screenWidth/2.0f;
	float posY = this->screenHeight/2.0f;

	this->vertices[0] = glm::vec4(posX,						posY - this->screenHeight,	0.0f, 1.0f);
    this->vertices[1] = glm::vec4(posX + this->screenWidth,	posY - this->screenHeight,	0.0f, 1.0f);
	this->vertices[2] = glm::vec4(posX + this->screenWidth,	posY,	0.0f, 1.0f);
    this->vertices[3] = glm::vec4(posX,						posY,	0.0f, 1.0f);
}

void OpenGLColorgradientWidget::setMode(int mode)
{
	m_mode = mode;
	updateGL();
}

void OpenGLColorgradientWidget::checkGLError()
{
	GLenum error = glGetError();
	if (error != GL_NO_ERROR)
	{
		std::cout << "_______________________________" << std::endl;
		std::cout << "OpenGLColorgradientWidget" << std::endl;

		switch (error)
		{
			case GL_INVALID_ENUM:
				cout << "GL: enum argument out of range." << endl;
				break;
			case GL_INVALID_VALUE:
				cout << "GL: Numeric argument out of range." << endl;
				break;
			case GL_INVALID_OPERATION:
				cout << "GL: Operation illegal in current state." << endl;
				break;
			case GL_INVALID_FRAMEBUFFER_OPERATION:
				cout << "GL: Framebuffer object is not complete." << endl;
				break;
			case GL_OUT_OF_MEMORY:
				cout << "GL: Not enough memory left to execute command." << endl;
				break;
			default:
				cout << "GL: Unknown error." << endl;
		}
		std::cout << "_______________________________" << std::endl;
	}
}