#pragma once

#include <GL/glew.h>
#include <QGLWidget>
//#include <QtOpenGL>
#include <QColor>
#include <QGLContext>
#include <QGLFormat>
#include <QMouseEvent>
#include <QTime>

#include "GLM/glm.hpp"
#include "SceneManager.h"
#include "Slicing.h"
#include "RayCasting.h"
#include "TransferPoints.h"

class OpenGLWidget : public QGLWidget
{
	Q_OBJECT

public:
	/** The constructor for the OpenGlWidget. */
	OpenGLWidget(QWidget* parent = 0);
	/** The destructor for the OpenGlWidget. */
	~OpenGLWidget();

public slots:
	/** 
	 * Sets the perspective according to the given Parameter.
	 * @param tabIndex The index of the current tab. 
	 */
	void setPerspective(int tabIndex);

	/** Generates a texture from the current volume. */
	void initSliceTexture();

	/** 
	 * Sets the x-coordinate for the slice.
	 * @param x The x-coordinate of the slice.
	 */
	void setXSlicing(int x);

	/** 
	 * Sets the y-coordinate for the slice.
	 * @param y The y-coordinate of the slice.
	 */
	void setYSlicing(int y);

	/** 
	 * Sets the z coordinate for the slice.
	 * @param z The z-coordinate of the slice.
	 */
	void setZSlicing(int z);

	/** 
	 * Sets the x-coordinate for clipping.
	 * @param x The x-coordinate.
	 */
	void setXClipping(int x);

	/** 
	 * Sets the y-coordinate for clipping.
	 * @param y The y-coordinate.
	 */
	void setYClipping(int y);

	/** 
	 * Sets the z coordinate for clipping.
	 * @param z The z-coordinate.
	 */
	void setZClipping(int z);

	/**
	 * Generates a 1D texture for the gives RGBA values.
	 * @param rgbaValues The RGBA values for the texture.
	 * @param textureSize The size of the texture.
	 */
	void generateTransferTexture(float* rgbaValues, int textureSize, int type);

	/** Deletes the current transfer texture. */
	void resetTransferTexture();

	/** 
	 * Activates or deactivates the light according to the given parameter.
	 * @param active True to activates the light otherwise false.
	 */
	void setLight(bool active);

	/** 
	 * Activates or deactivates the specular light according to the given parameter.
	 * @param active True to activates the specular light otherwise false.
	 */
	void setPhong(bool active);

	/** 
	 * Activates or deactivates the maximum intensity projection according to the given parameter.
	 * @param active True to activates the maximum intensity projection otherwise false.
	 */
	void setMip(bool active);

	/** 
	 * Activates or deactivates X-Ray (averaging) according to the given parameter.
	 * @param active True to activates X-Ray (averaging) otherwise false.
	 */
	void setAverage(bool active);

	/** 
	 * Activates or deactivates compositing according to the given parameter.
	 * @param active True to activates compositing otherwise false.
	 */
	void setCompositing(bool active);

	/**
	* Activates or deactivates adaptive - Stepsize according to the given parameter.
	* @param active True to activate adaptive Stepsize otherwise false.
	*/
	void setAdaptive(bool active);

	/**
	* Changed the Stepsize to the given parameter.
	*/
	void setStepSize(double step);

	/**
	* Changed the Isovalue to the given parameter.
	*/
	void changeIsoSurface(float val);

signals:
	/** 
	 * TF changed
	 * @param t .. t value
	 * @param s ... scalar value
	 */
	void valuesUpdated(glm::vec2* t, int s);

protected:
	void setValuesUpdated(glm::vec2* values, int size);

	/** Initialize GLEW and sets up an OpenGL 3.2 context. */
	void initializeGL();
	/** 
	 * Sets the new viewport an updates the projection matrix. 
	 * @param width The new width of the window, measured in pixel.
	 * @param height The new height of the window, measured in pixel.
	 */
	void resizeGL(int width,int height);
	/** Renders the scene. */
	void paintGL();

	/** Handles the mousePressEvents. */
	void mousePressEvent(QMouseEvent *event);
	/** Handles the mouseMoveEvent. */
	void mouseMoveEvent(QMouseEvent *event);
	/** Handles the mouseReleaseEvent. */
	void mouseReleaseEvent(QMouseEvent *event);
	/** Handles the wheelEvent. */
	void wheelEvent(QWheelEvent *event);

private:
	/** The width of the screen. */
	int screenWidth;
	/** The height of the screen. */
	int screenHeight;
	/** Boolean to set perspective or orthogonal view. */
	bool perspective;
	/** For render the slices in orthogonal view. */
	Slicing* slice;
	/** For render the 3D Volume. */
	RayCasting* rayCasting;
	/** The texturehandle for the transfer function texture. */
	GLuint transferFunctionTextureHandle;
	/** The texturehandle for the 2D transfer function texture. */
	GLuint transferFunctionTextureHandle2D;
	/** The texturehandle for the particle transfer function texture. */
	GLuint transferFunctionTextureParticleHandle;

	/** True if a volume move takes place. */
	bool moveVolume;
	/** True if a zoom takes place. */
	bool zoomVolume;
	/** The position of the mouse cursor at the last move operation. */
	glm::vec2 lastMovePos;
	/** The y-position of the mouse cursor at the last zoom operation. */
	int lastZoomYPos;

	/** Sets the projection matrix from the camera depending on the current window size. */
	void setCameraProjection();

	/** Basic OpenGL Error-Handling function. If encountered an error it prints an error-message to the console.*/
	void checkGLError();

	/** QTimer **/
	QTime timer;
	bool isStarted;

	int m_mode; 

	CudaArray<float4>* m_volumeBlock;
	CudaTexture<float4,3>* m_volumen;
	CudaTexture<float4,3>* m_pcd;
	CudaTexture<float4,2>* m_transferfunction;

	void OpenGLWidget::kMeans(glm::vec2** x, glm::vec2* m, float* mumCompForClass, float** g, int N, int K, int Iterations);
};