#pragma once
#include "RendererOpenGL.h"
#include "VertexGeometry.h"
#include "Primitives.h"
#include "Shader.h"
#include "ParticleSystemCuda.h"
#include "CudaTexture.h"

#include "GLM/gtx.hpp"
#include "GLM/gtc.hpp"
#include "GLM/gtx/type_ptr.hpp"

class RayCasting
{

public:
	/** The constructor. */
	RayCasting();
	/** The destructor. */
	~RayCasting();

	/** 
	 * Initialize the shaders an textures.
	 */
	void init();
	/** 
	 * Sets the new window size, calculates the vertices from the quad 
	 * and regenerates the textures for the give parameters. 
	 * @param width The new width of the window, measured in pixel.
	 * @param height The new height of the window, measured in pixel.
	 */
	void resize(int width, int height);
	/** 
	 * Renders the scene. 
	 */
	void render(float elapsedTime);

	/**
	 * Sets the texturehandle for the volume texture.
	 * @param tex The texture ID.
	 */
	void setVolumeTextureHandle(GLuint tex);
	/**
	 * Sets the texturehandle for the tranfer texture.
	 * @param tex The texture ID.
	 */
	void setTransferTextureHandle(GLuint tex);
	/**
	 * Sets the texturehandle for the 2D tranfer texture.
	 * @param tex The transfer texture 2D.
	 */
	void RayCasting::setTransferTexture2DHandle(GLuint tex);
	/**
	 * Sets the texturehandle for the tranfer texture for particles.
	 * @param tex The texture ID.
	 */
	void setTransferTextureParticleHandle(GLuint tex);
	/** 
	 * Activates or deactivates the transfer function according to the given parameter.
	 * @param active True to activates the transfer function otherwise false.
	 */
	void setTransferFunction(bool active);
	/**
	 * Sets the width, height and depth for the current volume.
	 * @param width The widht of the volume.
	 * @param height The height of the volume.
	 * @param depth The depth of the volume.
	 */
	void setVolumeSize(float width, float height, float depth);
	/**
	* Sets the value of the textureLoaded property.
	* @param loaded True if the texture of the volume is loaded, false otherwise.
	*/
	void setTextureLoaded(bool loaded);
	/** 
	 * Activates or deactivates the light according to the given parameter.
	 * @param active True to activates the light otherwise false.
	 */
	void setLight(bool active);
	/** 
	 * Activates or deactivates the specular light according to the given parameter.
	 * @param active True to activates the specular light otherwise false.
	 */
	void setPhong(bool active);
	/** 
	 * Activates or deactivates the maximum intensity projection according to the given parameter.
	 * @param active True to activates the maximum intensity projection otherwise false.
	 */
	void setMip(bool active);
	/** 
	 * Activates or deactivates X-Ray (averaging) according to the given parameter.
	 * @param active True to activates X-Ray (averaging) otherwise false.
	 */
	void setAverage(bool active);
	/** 
	 * Activates or deactivates compositing according to the given parameter.
	 * @param active True to activates compositing otherwise false.
	 */
	void setCompositing(bool active);
	/**
	* Activates or deactivates adaptive - Stepsize according to the given parameter.
	* @param active True to activate adaptive Stepsize otherwise false.
	*/
	void setAdaptive(bool active);
	/**
	* Sets the StepSize to the given parameter
	*/
	void setStepSize(double step);
	/** 
	 * Sets the x-coordinate for clipping.
	 * @param x The x-coordinate.
	 */
	void setXClipping(float x);
	/** 
	 * Sets the y-coordinate for clipping.
	 * @param y The y-coordinate.
	 */
	void setYClipping(float y);
	/** 
	 * Sets the z coordinate for clipping.
	 * @param z The z-coordinate.
	 */
	void setZClipping(float z);

	/**
	 * Sets the texturehandle for preprocessed PCD
	 * @param tex The texture 3D.
	 */
	void setPCDTextureHandle(GLuint tex);

	/**
	 * Change ISO Value
	 * @param val new iso value
	 **/
	void SetIsoValue(float val);

protected:

private:
	/** The width of the screen. */
	int screenWidth;
	/** The height of the screen. */
	int screenHeight;

	/** True if the volume texture is loaded.*/
	bool textureLoaded;

	/** The width of the current volume. */
	float volumeWidth;
	/** The height of the current volume. */
	float volumeHeight;
	/** The depth of the current volume. */
	float volumeDepth;
	/** The texturehandle for the volume texture. */
	GLuint volumeTextureHandle;
	/** The texturehandle for preprocessed PCD texture. */
	GLuint pcdTextureHandle;

	/** The texturehandle for the transfer function texture. */
	GLuint transferTextureHandle;
	/** The texturehandle for the 2D transfer function texture. */
	GLuint transferTextureHandle2D;
	/** The texturehandle for the particle texture. */
	GLuint tfParticleHandle;

	/** True if the transfer function is active.*/
	bool transferFunction;
	/** True if the light is active.*/
	bool light;
	/** True if the specular light is active.*/
	bool phong;
	/** True if the maximum intensity projection is active.*/
	bool mip;
	/** True if X-Ray is active.*/
	bool average;
	/** True if adaptive - Stepsize is active. */
	bool adaptive;
	/** Stepsize, indicates the distance between two sampling points */
	double step;

	/** The vertices for rendering the volume to the screen. */
	glm::vec4* quadVertices;

	/** The x-coordinate for clipping. */
	float xClipping;
	/** The y-coordinate for clipping. */
	float yClipping;
	/** The z-coordinate for clipping. */
	float zClipping;

	/** Bind all nessesary texture for ray casting. */
	void bindRayCastingTextures();
	/** Unbind all texture from ray casting. */
	void unbindRayCastingTextures();

	//for test only!!
	Shader* texturShader;

public:
	View* m_volumeBricks_front;
	View* m_volumeBricks_back;
//_________________
private:
	VertexGeometry* m_particle_geo;
	RendererOpenGL* m_renderer;

	ParticleSystemCuda* m_particle_cuda;

	FBO::FBOPtr fbo_back;
	FBO::FBOPtr fbo_front;
	FBO::FBOPtr fbo_hdr;
	View* m_bb_front;
	View* m_bb_back;
	View* m_raycasting;
	View* m_raycasting2D;
	View* m_mip;
	View* m_averaging;
	View* m_particle;
	View* m_hdr;
	View* m_cp_back;
	View* m_cp_front;
	View* m_cp2_back;
	View* m_cp2_front;
	
	View* m_contour;
	View* m_exposure;

	Camera* cam;

	unsigned int bspile_weights;
	unsigned int bspile_dev_weights;
public:
	VertexGeometry* m_plane_tmp;

public:
	int m_type;

	//COPY VOM CUDA SDK UM ZU TESTEN, OB DIE VOLUMENDATEN BENZT WERDEN KNNEN!!!
	void initPixelBuffer();
	void renderCuda();
	void display();
	float invViewMatrix[12];
	int width;
	int height;
	GLuint pbo;     // OpenGL pixel buffer object
	GLuint tex;     // OpenGL texture object
	struct cudaGraphicsResource *cuda_pbo_resource; // CUDA Graphics Resource (to transfer PBO)

	void AfterLoadingOfVolume(int width, int height, int depth);

	CudaTexture<float4,3>* m_volumen;
	CudaTexture<float4,2>* m_texture;

	CudaTexture<float4,2>* m_hg;
	CudaTexture<float4,2>* m_hg_dev;

	CudaTexture<float4,3>* m_pcd;
};