#pragma once
#include "Clustering.h"

#include <QObject>
#include <QRectF>
#include <QVector>
#include <QSizeF>
#include <QPen>
#include <QBrush>
#include <QPointF>
#include <QWidget>
#include <QColor>

#include "GLM/gtx.hpp"
#include "GLM/gtc.hpp"
#include "GLM/gtx/type_ptr.hpp"

#include "TPoint.h"

class TFObjects
{
public:
	int type;
	glm::vec2 size;
	glm::vec2 pos;
	glm::vec2 scale;
	int rotate;
	QColor color;
	QPainterPath path;

	glm::vec2 m;
	glm::mat2 c;
};

class TransferPoints : public QObject
{
	Q_OBJECT

public:
	enum PointShape {
        CircleShape,
        RectangleShape
    };

	enum LockType {
		LockToLeft   = 0x01,
		LockToRight  = 0x02,
		LockToTop    = 0x04,
		LockToBottom = 0x08
	};

	enum EventObjectType 
	{
		Selection		= 0,
		DrawPoint		= 1,
		DrawEllipse		= 2,
		DrawRectangle   = 3
	};

	enum SortType {
		NoSort,
		XSort,
		YSort
	};

	enum ConnectionType {
		NoConnection,
		LineConnection,
		CurveConnection
	};

	/** 
	*	Constructs a widget which is a child of widget, and the PointShape
	*/
	TransferPoints(QWidget *widget, PointShape shape);

	/**
	*	Reimplementation of eventFilter to filter the different events and handel the
	*	MouseEvents to set/delet/move Points in the TransferfunktionWidget
	*/
	bool eventFilter(QObject *object, QEvent *event);

	/** 
	*	Paints Transferpoints, connectionline, and Histogramm
	*/
	void paintPoints();

	/** 
	*	Returns a QRectF with the given size or size of the widget
	*/
	inline QRectF boundingRect() const;
	/** 
	*	Sets m_bounds to a given Size
	*/
	void setBoundingRect(const QRectF &boundingRect) { m_bounds = boundingRect; }

	/** 
	*	Returns a QVector with the points of the transferfunction
	*/
	const QVector<TPoint>& points() const;
	/** 
	*	Sets the points of the transferfunction
	*/
	void setPoints(const QVector<TPoint> &points);

	/** 
	*	Returns a QSizeF with the width and Height of the point
	*/
	QSizeF pointSize() const { return m_pointSize; }
	/** 
	*	Sets the size of a Point as QSizeF ( int width, int height )
	*/
	void setPointSize(const QSizeF &size) { m_pointSize = size; }
	
	/** 
	*	Returns the sortType which defines how the points are sortet
	*/
	SortType sortType() const { return m_sortType; }
	/** 
	*	Sets the sortType 
	*   #SortType
	*/
	void setSortType(SortType sortType) { m_sortType = sortType; }
	
	/** 
	*	Returns the connectionType wich defines how the Points are connected
	*	@warning { the ConnectionType only changed the apperance of the Line }
	*/
	ConnectionType connectionType() const { return m_connectionType; }
	/** 
	*	Sets the connectionType 
	*   #ConnectionType
	*	@warning { the ConnectionType only changed the apperance of the Line }
	*/
	void setConnectionType(ConnectionType connectionType) { m_connectionType = connectionType; }

	/** 
	*	Defines the apperance of the Line, with a QPen
	*/
	void setConnectionPen(const QPen &pen) { m_connectionPen = pen; }
	/** 
	*	Defines the apperance of the ShapeLine, with a QPen
	*/
	void setShapePen(const QPen &pen) { m_pointPen = pen; }
	/** 
	*	Defines the apperance of the Shape filling, with a QBrush
	*/
	void setShapeBrush(const QBrush &brush) { m_pointBrush = brush; }

	/** 
	*	Locks the "Point" to a Border 
	*	#LockType
	*/
	void setPointLock(int pos, LockType lock) { m_locks[pos] = lock; }

	/** 
	*	Sets whether the item is editable.
	*/
	void setEditable(bool editable) { m_editable = editable; }
	/** 
	*	Returns true if the item is editable
	*/
	bool editable() const { return m_editable; }


public slots:
	/** 
	*	Sets whether the item is enabled, an enabled widget handles keyboard and mouse events
	*/
	void setEnabled(bool enabled);

	/** 
	*	Disables widget input events if disable is true
	*/
	void setDisabled(bool disabled) { setEnabled(!disabled); }

	/** 
	*	Stores the color in p_color if the color has changed
	*/
	void changePointColor(QColor c) {p_color = c;}

	/** 
	*	Updates the histogramm
	*/
	void updateHistogram(double* histogram, int size);

	/**
	*	Update values
	*/
	void updateValues(glm::vec2* values, int size);

	/**
	* Modus of Transferfunction
	* 0... 1D TF
	* 1... 1D Gradient
	**/
	void updateMode(int mode);

	void DoClustering(int K);


	void updateActionObjectType(int type);
signals:

	/** 
	*	This signal is emitted when the points have changed
	*/
	void pointsChanged(int);
public:

	/** 
	*	Is called whenever the transferfunction has been modified, sort ans emit pointsChanged
	*/
	void firePointChange();

	std::vector<TFObjects> tf_objects;

private:
	/** 
	*	Returns a QRectF to define the area of the Point depending on the size
	*/
	inline QRectF pointBoundingRect(int i) const;

	/** 
	*	Returns a QRectF to define the area, so that the points don't overlap
	*/
	inline QRectF insertBoundingRect(int i) const;

	/** 
	*	Move currentPoint to a new Position 
	* @param i is the index of the Point which had been moved
	* @param newPos new Position of the Point
	*/
	void movePoint(int i, const TPoint &newPos, bool emitChange = true);
	
	/** Pointer to the widget */
	QWidget* m_widget;

	/** Stores the points if the transferfunction */
	QVector<TPoint>* m_points;
	
	/** The size of the widget in height and width */
	QRectF m_bounds;

	/** Verify the apperance of the points */
	PointShape m_shape;
	/** Verify how the points are sorted #SortType*/
	SortType m_sortType;
	/** Verify how the line between two points looks like #ConnectionType */
	ConnectionType m_connectionType;
	/** Stores the LockType of each Point */
	QVector<uint> m_locks;
	/** Verify the size of the points */
	QSizeF m_pointSize;
	/** The index of the point which is moved/changed... */
	int m_currentIndex;
	
	/** True if editable */	
	bool m_editable;
	/** True if enabled */
	bool m_enabled;
	
	/** Defines the apperance of the line of the point */
	QPen m_pointPen;
	/** Defines the filling of the points */
	QBrush m_pointBrush;
	/** Defines teh apperance of the connection Line */
	QPen m_connectionPen;
	/** The color of the points used to set m_pointPen and m_pointBrush */
	QColor p_color;

	/** The values of the histogramm */
	double* hist;
	/** The width of the histogramm */
	int histSize;
	//void drawHistogram(QPainter p);

	/**
	* Stores xy values (x... Density y... Gradient)
	*/
	glm::vec2* valuesX;
	// Stores Size
	int sizeX;

	//Stores Mode
	int m_mode;

	//Defines the current Operation for Actions
	int m_action_type;

	QPainterPath drawing_path;
	glm::vec2 start_rectangle;
	glm::vec2 act_rectangle;
	bool newAction;
};


inline QRectF TransferPoints::pointBoundingRect(int i) const
{
	int mode = m_mode;
	TPoint p = m_points[mode].at(i);
	double w = m_pointSize.width();
	double h = m_pointSize.height();
	double x = p.x() - w / 2;
	double y = p.y() - h / 2;
	return QRectF(x, y, w, h);
}

inline QRectF TransferPoints::insertBoundingRect(int i) const
{
	int mode = m_mode;
	TPoint p = m_points[mode].at(i);
	double w = m_pointSize.width() * 2.0;
	double h = m_pointSize.height() * 2.0;
	double x = p.x() - w / 2;
	double y = p.y() - h / 2;
	return QRectF(x, y, w, h);
}

inline QRectF TransferPoints::boundingRect() const
{
	 if (m_bounds.isEmpty())
		 return m_widget->rect();
	 else
		 return m_bounds;
};