#pragma once
#include "particleCuda.cuh"

#include <cuda.h>
#include <cuda_runtime.h>
#include <cmath>

typedef double FloatType;

//Speichert Gradient (x,yz) und Dichtewert (w)
texture<float4,3>  volume_tex;
texture<float4,3>  pcd_tex;
texture<float4,2>  transferfunction_tex;
texture<float4,2>  bspline_hg_tex;
texture<float4,2>  bspline_hg_dev_tex;


/*
 *	Singular Value Decomposition for CUDA kernels
 *	No dependencies except cmath and CUDA. 
 *	
 *	Written by: http://www.denniskempin.com
 *	And many thanks to: http://www.melax.com/diag
 */

/*
 * Small helper class for a 3D vector. 
 * Allows elements to be accessed as x, y, z 
 * Or as v[0], v[1], v[2].
 */
class Vector3
{
public:
	union 
	{
		FloatType v[3];
		struct
		{
			FloatType x, y, z;
		};
	};

	__device__ __host__ FloatType& operator[](const int idx)
	{
		return v[idx];
	}

	__device__ __host__ Vector3() : x(0), y(0), z(0)
	{

	}
	__device__ __host__ Vector3(FloatType x, FloatType y, FloatType z) : x(x), y(y), z(z)
	{

	}
};

/*
 * 3x3 Matrix class. 
 * Allows elements to be accessed as m00, m01, ... 
 * Or as v[0], v[1], ...
 * Elements are stored in row major order
 * m00 = v[0]; m01 = v[1]; ...
 */
struct Matrix3x3 
{
	union 
	{
		FloatType v[9];
		struct
		{
			FloatType m00, m01, m02;
			FloatType m10, m11, m12;
			FloatType m20, m21, m22;
		};
	};

	__device__ __host__ FloatType& operator[](const int idx)
	{
		return v[idx];
	}

	/*
	 *	Identity constructor
	 */
	__device__ __host__ Matrix3x3() :
		m00(1), m01(0), m02(0), 
		m10(0), m11(1), m12(0), 
		m20(0), m21(0), m22(1)

	{

	}

	/*
	 *	set elements constructor
	 */
	__device__ __host__ Matrix3x3(FloatType m00, FloatType m01, FloatType m02, FloatType m10, FloatType m11, FloatType m12, FloatType m20, FloatType m21, FloatType m22) :
		m00(m00), m01(m01), m02(m02), 
		m10(m10), m11(m11), m12(m12), 
		m20(m20), m21(m21), m22(m22)

	{

	}

	/*
	 *	Copy constructor
	 */
	__device__ __host__ Matrix3x3(const Matrix3x3& M) :
	m00(M.m00), m01(M.m01), m02(M.m02), 
		m10(M.m10), m11(M.m11), m12(M.m12), 
		m20(M.m20), m21(M.m21), m22(M.m22)

	{

	}

	/*
	 *	Element-wise minus operator
	 */
	__device__ __host__ Matrix3x3 operator-(Matrix3x3& B) {
		Matrix3x3& A = *this;
		Matrix3x3 C; 

		for(int i=0; i<9; ++i) C[i] = A[i] - B[i];

		return C;
	}

	/*
	 *	Calculate sum of all squared elements
	 */
	__device__ __host__ FloatType SquaredSum() {
		Matrix3x3& A = *this;
		FloatType weight = 0;

		for(int i=0; i<9; ++i) weight += A[i]*A[i];

		return weight;
	}

	/*
	 *	Matrix*Matrix multiplication
	 */
	__device__ __host__ Matrix3x3 operator*(const Matrix3x3& A) const {
		const Matrix3x3& B = *this;
		Matrix3x3 C; 

		C.m00 = A.m00*B.m00 + A.m10*B.m01 + A.m20*B.m02;
		C.m10 = A.m00*B.m10 + A.m10*B.m11 + A.m20*B.m12;
		C.m20 = A.m00*B.m20 + A.m10*B.m21 + A.m20*B.m22;

		C.m01 = A.m01*B.m00 + A.m11*B.m01 + A.m21*B.m02;
		C.m11 = A.m01*B.m10 + A.m11*B.m11 + A.m21*B.m12;
		C.m21 = A.m01*B.m20 + A.m11*B.m21 + A.m21*B.m22;

		C.m02 = A.m02*B.m00 + A.m12*B.m01 + A.m22*B.m02;
		C.m12 = A.m02*B.m10 + A.m12*B.m11 + A.m22*B.m12;
		C.m22 = A.m02*B.m20 + A.m12*B.m21 + A.m22*B.m22;

		return C;
	}

	/*
	 *	Matrix*Vector multiplication
	 */
	__device__ __host__ Vector3 operator*(const Vector3& B) const {
		const Matrix3x3& A = *this;
		Vector3 C; 

		C.x = A.m00*B.x + A.m01*B.y + A.m02*B.z;
		C.y = A.m10*B.x + A.m11*B.y + A.m12*B.z;
		C.z = A.m20*B.x + A.m21*B.y + A.m22*B.z;

		return C;
	}

	/*
	 *	Matrix*Skalar multiplication
	 */
	__device__ __host__ Matrix3x3 operator*(const double& B) const {
		const Matrix3x3& A = *this;
		Matrix3x3 C; 

		C.m00 = A.m00*B;
		C.m01 = A.m01*B;
		C.m02 = A.m02*B;
		C.m10 = A.m10*B;
		C.m11 = A.m11*B;
		C.m12 = A.m12*B;
		C.m20 = A.m20*B;
		C.m21 = A.m21*B;
		C.m22 = A.m22*B;

		return C;
	}

	/*
	 *	3x3 Determinant
	 */
	__device__ __host__ FloatType Determinant() {
		FloatType out;

		out  = m00*m11*m22;
		out += m01*m12*m20;
		out += m02*m10*m21;
		
		out -= m20*m11*m02;
		out -= m10*m01*m22;
		out -= m00*m21*m12;

		return out;
	}
};


/*
 *	Returns a transposed version of the matrix
 */
__device__ __host__ inline Matrix3x3 Transposed(Matrix3x3 A) {
	Matrix3x3 out;

	out.m00 = A.m00;
	out.m01 = A.m10;
	out.m02 = A.m20;

	out.m10 = A.m01;
	out.m11 = A.m11;
	out.m12 = A.m21;

	out.m20 = A.m02;
	out.m21 = A.m12;
	out.m22 = A.m22;

	return out;
}

/*
 * Quaternion helper class
 * Allows elements to be accessed as x, y, z, w
 * Or as v[0], v[1], v[2], v[3].
 */
struct Quaternion 
{
	union 
	{
		FloatType v[4];
		struct
		{
			FloatType x, y, z, w;
		};
	};

	__device__ __host__ FloatType& operator[](const int idx)
	{
		return v[idx];
	}

	/*
	 *	Identity
	 */
	__device__ __host__ Quaternion() : x(0), y(0), z(0), w(1)
	{

	}

	/*
	 *	Basic Constructor
	 */
	__device__ __host__ Quaternion(FloatType x, FloatType y, FloatType z, FloatType w) : x(x), y(y), z(z), w(w)
	{

	}

	/*
	 *	Quaternion concatenation operator
	 */
	__device__ __host__ Quaternion operator*(const Quaternion& b)
	{
		Quaternion tmp;

		tmp.w = (b.w * w) - (b.x * x) - (b.y * y) - (b.z * z);
		tmp.x = (b.w * x) + (b.x * w) + (b.y * z) - (b.z * y);
		tmp.y = (b.w * y) + (b.y * w) + (b.z * x) - (b.x * z);
		tmp.z = (b.w * z) + (b.z * w) + (b.x * y) - (b.y * x);

		return tmp;
	}

	/**
	 * Normalize the quaternion
	 */
	__device__ __host__ void Normalize()
	{
		FloatType lengthInv = 1.0/sqrt(x*x + y*y + z*z + w*w);
		x *= lengthInv;
		y *= lengthInv;
		z *= lengthInv;
		w *= lengthInv;
	}

	/**
	 * Returns matrix representation of quaternion as a rotation description
	 */
	__device__ __host__ Matrix3x3 GetMatrix()
	{
		Matrix3x3 out;
		
		out.m00 = 1 - 2 * ( y*y + z*z );
		out.m01 =     2 * ( x*y - z*w );
		out.m02 =     2 * ( x*z + y*w );

		out.m10 =     2 * ( x*y + z*w );
		out.m11 = 1 - 2 * ( x*x + z*z );
		out.m12 =     2 * ( y*z - x*w );

		out.m20 =     2 * ( x*z - y*w );
		out.m21 =     2 * ( y*z + x*w );
		out.m22 = 1 - 2 * ( x*x + y*y );
		
		return out;
	}
};

/*
 * Returns Quaternion that diagonalizes A. 
 * A must be symmetric.
 */
__device__ __host__ inline Quaternion Diagonalize(const Matrix3x3& A)
{
	// Diagonal matrix D = Q * A * FloatTyperanspose(Q);  and  A = QFloatType*D*Q
	// FloatTypehe rows of q are the eigenvectors D's diagonal is the eigenvalues
	// As per 'row' convention if float3x3 Q = q.getmatrix(); then v*Q = q*v*conj(q)
	int i;
	Quaternion q = Quaternion(0, 0, 0, 1);
	for(i=0; i<100; i++)
	{
		Matrix3x3 Q  = q.GetMatrix(); // v*Q == q*v*conj(q)
		Matrix3x3 D  = Transposed(Q) * A * Q;  // A = Q^FloatType*D*Q

		Vector3 diag(D.m00, D.m11, D.m22);
		Vector3 offdiag(D.m12, D.m02, D.m01); // elements not on the diagonal
		Vector3 om(fabs(offdiag.x), fabs(offdiag.y), fabs(offdiag.z)); // mag of each offdiag elem
		
		int k = (om.x > om.y && om.x > om.z) ? 0 : (om.y > om.z) ? 1 : 2; // index of largest element of offdiag
		int k1 = (k+1)%3;
		int k2 = (k+2)%3;
		if(offdiag[k]==0.0f) break;  // diagonal already
		
		FloatType thet = (diag[k2]-diag[k1])/(2.0f*offdiag[k]);
		FloatType sgn = (thet > 0.0f) ? 1.0f : -1.0f;
		thet    *= sgn; // make it positive
		FloatType t = sgn /(thet +((thet < 1.E6f)?sqrt(thet*thet+1.0f):thet)) ; // sign(FloatType)/(|FloatType|+sqrt(FloatType^2+1))
		FloatType c = 1.0f/sqrt(t*t+1.0f); //  c= 1/(t^2+1) , t=s/c 
		if(c==1.0f) break;  // no room for improvement - reached machine precision.
		
		Quaternion jr(0,0,0,0); // jacobi rotation for this iteration.
		jr[k] = sgn*sqrt((1.0f-c)/2.0f);  // using 1/2 angle identity sin(a/2) = sqrt((1-cos(a))/2)  
		jr[k] *= -1.0f; // since our quat-to-matrix convention was for v*M instead of M*v
		jr.w  = sqrt(1.0f - jr[k]*jr[k]);
		if(jr.w==1.0f) break; // reached limits of floating point precision
		q =  q*jr;  
		q.Normalize();
	} 
	return q;
}

/*
 * Returns matrix containing all eigenvectors
 * Eigenvectors are sorted by the magnitude of their eigenvalues. 
 */
__device__ __host__ inline Matrix3x3 EigenvectorDecomposition(const Matrix3x3& A, Vector3 &eigValues)
{
	// Calculate eigenvector matrix
	Matrix3x3 Q = Diagonalize(A).GetMatrix();

	// Calculate diagonal matrix. Diagonal entries contain eigenvalues.
	Matrix3x3 D = Transposed(Q) * A * Q;
	Vector3 diag = Vector3(fabs(D.m00), fabs(D.m11), fabs(D.m22));
	Vector3 diag_2 = Vector3(D.m00, D.m11, D.m22);

	// Find index of biggest eigenvalue-
	int ev1 = (diag[0] > diag[1] && diag[0] > diag[2]) ? 0 : (diag[1] > diag[2]) ? 1 : 2;
	
	// Put last two eigenvalues into order.
	int ev2 = (ev1+1)%3;
	int ev3 = (ev1+2)%3;
	if(diag[ev3] > diag[ev2])
	{
		int tmp = ev2;
		ev2 = ev3;
		ev3 = tmp;
	}

	// Reorder matrix to start with the eigenvector of the biggest eigenvalue
	Matrix3x3 R;
	R.m00 = Q[0*3+ev1];
	R.m10 = Q[1*3+ev1];
	R.m20 = Q[2*3+ev1];

	R.m01 = Q[0*3+ev2];
	R.m11 = Q[1*3+ev2];
	R.m21 = Q[2*3+ev2];

	R.m02 = Q[0*3+ev3];
	R.m12 = Q[1*3+ev3];
	R.m22 = Q[2*3+ev3];
	
	eigValues[0] = diag_2[ev1];
	eigValues[1] = diag_2[ev2];
	eigValues[2] = diag_2[ev3];
	

	return R;
}

/*
 *	Calculates U and V matrix for a singular value decomposition
 *  A = U * S * VFloatType
 *  http://en.wikipedia.org/wiki/Singular_value_decomposition
 */
__device__ __host__ inline void SVD(const Matrix3x3& A, Matrix3x3& U, Matrix3x3& V)
{
	Matrix3x3 R;

	// Create square symmetric matrices
	Matrix3x3 AAFloatType = A * Transposed(A);
	Matrix3x3 AFloatTypeA = Transposed(A) * A;
	Vector3 eigenVectorU;
	Vector3 eigenVectorV;

	// U and V contain the eigenvalues of AAFloatType and AFloatTypeA
	U = EigenvectorDecomposition(AAFloatType, eigenVectorU);
	V = EigenvectorDecomposition(AFloatTypeA, eigenVectorV);
	
	// Calculate S diagonal matrix.
	Matrix3x3 S = Transposed(U) * A * V;

	// Singular values have to be positive. 
	// If one of the singular values is negative we have to
	// invert the corresponding column in U or V. I chose U. 
	if(S.m00 < 0) 
	{
		U.m00 = -U.m00;
		U.m10 = -U.m10;
		U.m20 = -U.m20;
	}

	if(S.m11 < 0) 
	{
		U.m01 = -U.m01;
		U.m11 = -U.m11;
		U.m21 = -U.m21;
	}

	if(S.m22 < 0) 
	{
		U.m02 = -U.m02;
		U.m12 = -U.m12;
		U.m22 = -U.m22;
	}
}

/*
 *  Polar decompositon based on singular value decomposition. Q = U*FloatTyperansposed(V)
 */
__device__ __host__ inline Matrix3x3 PolarDecomposition(const Matrix3x3& A)
{
	Matrix3x3 U, V;
    SVD(A, U, V);
    return U*Transposed(V);
}



/*
Diese Funktion unterteilt das Volumen in Bricks. Fr jeden Brick wird anschlieend der grte und kleine Wert gespeichert.

Im Shader "volumeBricks" wird im VertexShader Diese Werte werden anschlieend gegen die Transferfunktion gecullt. Falls es fr die Transferfunktion nicht von Bedeutung ist,
sprich fr den min, max Bereich Opacity 0 zugeordnet ist, wird der Block gecullt.

Im Geometry Shader werden dann anschlieend die Blocks erstellt und gerendert. Dadurch wird beim Raycasting Bereich mit geringer Bedeutung 
(sprich: ohne Einfluss auf das Ergbnis) ausgelassen. 

Mehr Infos unter: Object-Order Empty Space Skipping
*/

__global__ void CalcVolBlocks(float4* block, cudaExtent res, cudaExtent blocks_size, cudaExtent blocks_num, int bx) 
{
	int x = threadIdx.x + (blockIdx.x % bx) * blockDim.x;
    int y = threadIdx.y + blockIdx.y * blockDim.y;
	int z = threadIdx.z + (blockIdx.x / bx) * blockDim.x;

	float3 dim = make_float3(res.width, res.height, res.depth);
	
	float2 min_max = make_float2(999.0f, -999.0);
	if (x<blocks_num.width && y<blocks_num.height && z<blocks_num.depth && (z*blocks_num.width*blocks_num.height + y*blocks_num.width + x) < (blocks_num.width*blocks_num.height*blocks_num.depth)) 
	{
		for(int i = 0; i < blocks_size.width; i++)
		{
			for(int j = 0; j < blocks_size.height; j++)
			{
				for(int k = 0; k < blocks_size.depth; k++)
				{
					float data = tex3D(	volume_tex, 
										((float)(x*blocks_size.width)+i)/dim.x, 
										((float)(y*blocks_size.height)+j)/dim.y, 
										((float)(z*blocks_size.depth)+k)/dim.z).w;
					min_max.x = min(min_max.x, data);
					min_max.y = max(min_max.y, data);
				}
			}
		}
		
		block[z*blocks_num.width*blocks_num.height + y*blocks_num.width + x] = make_float4(z*blocks_num.width*blocks_num.height + y*blocks_num.width + x, min_max.x, min_max.y,1.f);
	}
}

extern "C" void CalcVolumeBlockCuda(float4* block, cudaExtent res, cudaExtent blocks_size, cudaExtent blocks_num)
{
	int nThreads = 8;
	dim3 threads;
	threads.x = nThreads;
	threads.y = nThreads;
	threads.z = nThreads;
	
	dim3 blocks;
	blocks.x = (blocks_num.width/nThreads + (!(blocks_num.width%nThreads)?0:1))
			 * (blocks_num.depth/nThreads + (!(blocks_num.depth%nThreads)?0:1));
	blocks.y = blocks_num.height/nThreads + (!(blocks_num.height%nThreads)?0:1);
	blocks.z = 1;
	
	int bx = ceil((float)blocks_num.width/(float)nThreads);
		
	CalcVolBlocks<<<blocks,threads>>>(block, res, blocks_size, blocks_num, bx);
}

extern "C" void initVolumenTexture(cudaArray* volArray, cudaExtent volumeSize)
{
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	volume_tex.normalized = true;                      // access with normalized texture coordinates
    volume_tex.filterMode = cudaFilterModeLinear;      // linear interpolation
    volume_tex.addressMode[0] = cudaAddressModeClamp;  // clamp texture coordinates
    volume_tex.addressMode[1] = cudaAddressModeClamp;
	volume_tex.addressMode[2] = cudaAddressModeClamp;

	cutilSafeCall(cudaBindTextureToArray(volume_tex, volArray, channelDesc));
}

extern "C" void initPCDTexture(cudaArray* volArray, cudaExtent volumeSize)
{
	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	pcd_tex.normalized = true;                      // access with normalized texture coordinates
    pcd_tex.filterMode = cudaFilterModeLinear;      // linear interpolation
    pcd_tex.addressMode[0] = cudaAddressModeClamp;  // clamp texture coordinates
    pcd_tex.addressMode[1] = cudaAddressModeClamp;
	pcd_tex.addressMode[2] = cudaAddressModeClamp;

	cutilSafeCall(cudaBindTextureToArray(pcd_tex, volArray, channelDesc));
}

extern "C" void BindVolumeCuda(cudaArray* volume_array)
{
	cudaChannelFormatDesc descFloat3d = cudaCreateChannelDesc<float4>();
	cutilSafeCall(cudaBindTextureToArray(volume_tex, volume_array, descFloat3d));
}

extern "C" void UnbindVolumeCuda()
{
	cutilSafeCall(cudaUnbindTexture(volume_tex));
}

extern "C" void BindPCDCuda(cudaArray* volume_array)
{
	cudaChannelFormatDesc descFloat3d = cudaCreateChannelDesc<float4>();
	cutilSafeCall(cudaBindTextureToArray(pcd_tex, volume_array, descFloat3d));
}

extern "C" void UnbindPCDCuda()
{
	cutilSafeCall(cudaUnbindTexture(pcd_tex));
}





//__________________________________
// Dieser Bereich wird nur zum Testen bentigt, ob die Texturen in Cuda lesbar sind.
// Wurde 1zu1 vom SDK bernommen.
// Aber so funktioniert Raycasting mit CUDA. 

typedef struct {
    float4 m[3];
} float3x4;

typedef struct {
    float4 m[4];
} float4x4;

typedef struct {
    float3 m[3];
} float3x3;

__constant__ float3x4 c_invViewMatrix;
__constant__ float4x4 c_modelViewMatrix;

struct Ray {
	float3 o;	// origin
	float3 d;	// direction
};

__device__ float3 mul(const float3x4 &M, const float3 &v)
{
    float3 r;
    r.x = dot(v, make_float3(M.m[0]));
    r.y = dot(v, make_float3(M.m[1]));
    r.z = dot(v, make_float3(M.m[2]));
    return r;
}

__device__ float3x3 mul(const float3x3 &v, const float3x3 &M)
{
    float3x3 r;

    r.m[0].x = (v.m[0].x * M.m[0].x) + (v.m[0].y * M.m[1].x) + (v.m[0].z * M.m[2].x);
    r.m[0].y = (v.m[0].x * M.m[0].y) + (v.m[0].y * M.m[1].y) + (v.m[0].z * M.m[2].y);
    r.m[0].z = (v.m[0].x * M.m[0].z) + (v.m[0].y * M.m[1].z) + (v.m[0].z * M.m[2].z);
	
	r.m[1].x = (v.m[1].x * M.m[0].x) + (v.m[1].y * M.m[1].x) + (v.m[1].z * M.m[2].x);
    r.m[1].y = (v.m[1].x * M.m[0].y) + (v.m[1].y * M.m[1].y) + (v.m[1].z * M.m[2].y);
    r.m[1].z = (v.m[1].x * M.m[0].z) + (v.m[1].y * M.m[1].z) + (v.m[1].z * M.m[2].z);

	r.m[2].x = (v.m[2].x * M.m[0].x) + (v.m[2].y * M.m[1].x) + (v.m[2].z * M.m[2].x);
    r.m[2].y = (v.m[2].x * M.m[0].y) + (v.m[2].y * M.m[1].y) + (v.m[2].z * M.m[2].y);
    r.m[2].z = (v.m[2].x * M.m[0].z) + (v.m[2].y * M.m[1].z) + (v.m[2].z * M.m[2].z);

    return r;
}

__device__ float det(float3x3 mat) 
{
	return	(mat.m[0].x * mat.m[1].y * mat.m[2].z) + (mat.m[0].y * mat.m[1].z * mat.m[2].x) + (mat.m[0].z * mat.m[1].x * mat.m[2].y) - 
			(mat.m[0].z * mat.m[1].y * mat.m[2].x) - (mat.m[0].y * mat.m[1].x * mat.m[2].z) - (mat.m[0].x * mat.m[1].z * mat.m[2].y);
}

__device__ float3x3 adjugate(float3x3 mat) 
{
	float3x3 mat2;
	mat2.m[0].x = (mat.m[1].y*mat.m[2].z)-(mat.m[1].z*mat.m[2].y);
	mat2.m[0].y =-(mat.m[0].y*mat.m[2].z)-(mat.m[0].z*mat.m[2].y);
	mat2.m[0].z = (mat.m[0].y*mat.m[1].z)-(mat.m[0].z*mat.m[1].y);

	mat2.m[1].x =-(mat.m[1].x*mat.m[2].z)-(mat.m[1].z*mat.m[2].x);
	mat2.m[1].y = (mat.m[0].x*mat.m[2].z)-(mat.m[0].z*mat.m[2].x);
	mat2.m[1].z =-(mat.m[0].x*mat.m[1].z)-(mat.m[0].z*mat.m[1].x);

	mat2.m[2].x = (mat.m[1].x*mat.m[2].y)-(mat.m[1].y*mat.m[2].x);
	mat2.m[2].y =-(mat.m[0].x*mat.m[2].y)-(mat.m[0].y*mat.m[2].x);
	mat2.m[2].z = (mat.m[0].x*mat.m[1].y)-(mat.m[0].y*mat.m[1].x);
	return mat2;
}

__device__ float3x3 transpose(float3x3 mat) 
{
	float3x3 mat2;
	mat2.m[0].x = mat.m[0].x;
	mat2.m[0].y = mat.m[1].x;
	mat2.m[0].z = mat.m[2].x;

	mat2.m[1].x = mat.m[0].y;
	mat2.m[1].y = mat.m[1].y;
	mat2.m[1].z = mat.m[2].y;

	mat2.m[2].x = mat.m[0].z;
	mat2.m[2].y = mat.m[1].z;
	mat2.m[2].z = mat.m[2].z;
	return mat2;
}

__device__ float3x3 mul(const float3x3 &M, const float d)
{
    float3x3 r;

    r.m[0].x = M.m[0].x/d;
    r.m[0].y = M.m[0].y/d;
    r.m[0].z = M.m[0].z/d;
	
	r.m[1].x = M.m[1].x/d;
    r.m[1].y = M.m[1].y/d;
    r.m[1].z = M.m[1].z/d;

	r.m[2].x = M.m[2].x/d;
    r.m[2].y = M.m[2].y/d;
    r.m[2].z = M.m[2].z/d;

    return r;
}

__device__ float3x3 minus(float3x3 matrix, float3x3 C)
{
	matrix.m[0].x -= C.m[0].x;
	matrix.m[0].y -= C.m[0].y;
	matrix.m[0].z -= C.m[0].z;

	matrix.m[1].x -= C.m[1].x;
	matrix.m[1].y -= C.m[1].y;
	matrix.m[1].z -= C.m[1].z;

	matrix.m[2].x -= C.m[2].x;
	matrix.m[2].y -= C.m[2].y;
	matrix.m[2].z -= C.m[2].z;
	return matrix;
}
 
__device__ float3x3 identity3x3(const float d)
{
    float3x3 r;

    r.m[0].x = d;
    r.m[0].y = 0.0;
    r.m[0].z = 0.0;
	r.m[1].x = 0.0;
    r.m[1].y = d;
    r.m[1].z = 0.0;
	r.m[2].x = 0.0;
    r.m[2].y = 0.0;
    r.m[2].z = d;

    return r;
}

__device__ float3x3 identity3x3(const float d1, const float d2)
{
    float3x3 r;

    r.m[0].x = d1;
    r.m[0].y = 0.0;
    r.m[0].z = 0.0;
	r.m[1].x = 0.0;
    r.m[1].y = d2;
    r.m[1].z = 0.0;
	r.m[2].x = 0.0;
    r.m[2].y = 0.0;
    r.m[2].z = 0.0;

    return r;
}

__device__ float3x3 inverse(float3x3 mat)
{
	float d = det(mat);
	float3x3 adj = adjugate(mat);
	adj = mul(adj, 1.0/d);
	return adj;
}

__device__ float3 calcEigenValue(float3 gradient, float3x3 matrix)
{
	float3 norm = gradient;
	float3 n = -normalize(norm);

	float3x3 H = matrix;

	float3x3 pn;
	pn.m[0].x = n.x;
	pn.m[1].x = n.y;
	pn.m[2].x = n.z;

	float3x3 id = identity3x3(1.0);
	float3x3 tpn = transpose(pn);

	float3x3 helper = mul(tpn,pn);
	float3x3 P = minus(id, helper);
	float3x3 C = mul(mul(P,H), -1.0/length(norm));

	return make_float3(C.m[0].x, C.m[1].y, C.m[2].z);
}

__device__ float3x3 calcEigenVector(float3 gradient, float3x3 matrix)
{
	float3 norm = gradient;
	float3 n = -normalize(norm);	

	float3x3 H = matrix;

	float3x3 pn;
	pn.m[0].x = n.x;
	pn.m[1].x = n.y;
	pn.m[2].x = n.z;

	float3x3 id = identity3x3(1.0);
	float3x3 tpn = transpose(pn);

	float3x3 helper = mul(tpn,pn);
	float3x3 P = minus(id, helper);
	
	float3x3 C = mul(mul(P,H), -1.0/length(norm));
	float3x3 tmp = minus(matrix,C);
	float3x3 vector = inverse(tmp);

	return vector;
}

__device__ float4 mul(const float3x4 &M, const float4 &v)
{
    float4 r;
    r.x = dot(v, M.m[0]);
    r.y = dot(v, M.m[1]);
    r.z = dot(v, M.m[2]);
    r.w = 1.0f;
    return r;
}

__device__ float4 mul(const float4x4 &M, const float4 &v)
{
    float4 r;

    r.x = dot(v, M.m[0]);
    r.y = dot(v, M.m[1]);
    r.z = dot(v, M.m[2]);
    r.w = dot(v, M.m[3]);

    return r;
}

__device__ int intersectBox(Ray r, float3 boxmin, float3 boxmax, float *tnear, float *tfar)
{
    // compute intersection of ray with all six bbox planes
    float3 invR = make_float3(1.0f) / r.d;
    float3 tbot = invR * (boxmin - r.o);
    float3 ttop = invR * (boxmax - r.o);

    // re-order intersections to find smallest and largest on each axis
    float3 tmin = fminf(ttop, tbot);
    float3 tmax = fmaxf(ttop, tbot);

    // find the largest tmin and the smallest tmax
    float largest_tmin = fmaxf(fmaxf(tmin.x, tmin.y), fmaxf(tmin.x, tmin.z));
    float smallest_tmax = fminf(fminf(tmax.x, tmax.y), fminf(tmax.x, tmax.z));

	*tnear = largest_tmin;
	*tfar = smallest_tmax;

	return smallest_tmax > largest_tmin;
}

__device__ uint rgbaFloatToInt(float4 rgba)
{
    rgba.x = __saturatef(rgba.x);   // clamp to [0.0, 1.0]
    rgba.y = __saturatef(rgba.y);
    rgba.z = __saturatef(rgba.z);
    rgba.w = __saturatef(rgba.w);
    return (uint(rgba.w*255)<<24) | (uint(rgba.z*255)<<16) | (uint(rgba.y*255)<<8) | uint(rgba.x*255);
}

__global__ void d_render(uint *d_output, uint imageW, uint imageH, float density, float brightness, float transferOffset, float transferScale)
{
    const int maxSteps = 500;
    const float tstep = 0.01f;
    const float opacityThreshold = 0.95f;
    const float3 boxMin = make_float3(-1.0f, -1.0f, -1.0f);
    const float3 boxMax = make_float3(1.0f, 1.0f, 1.0f);

	uint x = blockIdx.x*blockDim.x + threadIdx.x;
    uint y = blockIdx.y*blockDim.y + threadIdx.y;
    if ((x >= imageW) || (y >= imageH)) return;

    float u = (x / (float) imageW)*2.0f-1.0f;
    float v = (y / (float) imageH)*2.0f-1.0f;

    // calculate eye ray in world space
    Ray eyeRay;
    eyeRay.o = make_float3(mul(c_invViewMatrix, make_float4(0.0f, 0.0f, 0.0f, 1.0f)));
    eyeRay.d = normalize(make_float3(u, v, -2.0f));
    eyeRay.d = mul(c_invViewMatrix, eyeRay.d);

    // find intersection with box
	float tnear, tfar;
	int hit = intersectBox(eyeRay, boxMin, boxMax, &tnear, &tfar);
    if (!hit) return;
	if (tnear < 0.0f) tnear = 0.0f;     // clamp to near plane

    // march along ray from front to back, accumulating color
    float4 sum = make_float4(0.0f);
    float t = tnear;
    float3 pos = eyeRay.o + eyeRay.d*tnear;
    float3 step = eyeRay.d*tstep;

    for(int i=0; i<maxSteps; i++) {
        // read from 3D texture
        // remap position to [0, 1] coordinates
        float sample = tex3D(volume_tex, pos.x*0.5f+0.5f, pos.y*0.5f+0.5f, pos.z*0.5f+0.5f).w;

        // lookup in transfer function texture
        float4 col = tex2D(transferfunction_tex, (sample*transferScale)+transferOffset,0.5);
	    //float4 col = make_float4(sample);
        col.w *= density;

        // pre-multiply alpha
        col.x *= col.w;
        col.y *= col.w;
        col.z *= col.w;
        // "over" operator for front-to-back blending
        sum = sum + col*(1.0f - sum.w);

        // exit early if opaque
        if (sum.w > opacityThreshold)
            break;

        t += tstep;
        if (t > tfar) break;

        pos += step;
    }
    sum *= brightness;

    // write output color
    d_output[y*imageW + x] = rgbaFloatToInt(sum);
}

extern "C" void render_kernel(dim3 gridSize, dim3 blockSize, uint *d_output, uint imageW, uint imageH, float density, float brightness, float transferOffset, float transferScale)
{
	d_render<<<gridSize, blockSize>>>( d_output, imageW, imageH, density, brightness, transferOffset, transferScale);
}

extern "C"
void copyInvViewMatrix(float *invViewMatrix, size_t sizeofMatrix)
{
    cutilSafeCall( cudaMemcpyToSymbol(c_invViewMatrix, invViewMatrix, sizeofMatrix) );
}

extern "C" void initTransferFunctionTexture(cudaArray* tfArray)
{
//BIE NICHT NORMALISIERTEN MODUS MUSS MAN 0.5 0.5 hinzufgen um auf die korrekten Arrays zuzugreifen
//BEI NORMALISIERTEN WOHL AUCH. 

	cudaChannelFormatDesc channelDesc = cudaCreateChannelDesc<float4>();

	transferfunction_tex.normalized = true;                      // access with normalized texture coordinates
    transferfunction_tex.filterMode = cudaFilterModeLinear;      // linear interpolation
    transferfunction_tex.addressMode[0] = cudaAddressModeClamp ;  // clamp texture coordinates
	transferfunction_tex.addressMode[1] = cudaAddressModeClamp ;  // clamp texture coordinates

	cutilSafeCall(cudaBindTextureToArray(transferfunction_tex, tfArray, channelDesc));
}

extern "C" void BindTransferfunctionCuda(cudaArray* trans_array)
{
	cudaChannelFormatDesc descFloat = cudaCreateChannelDesc<float4>();
	cutilSafeCall(cudaBindTextureToArray(transferfunction_tex, trans_array, descFloat));
}

extern "C" void UnbindTransferfunctionCuda()
{
	cutilSafeCall(cudaUnbindTexture(transferfunction_tex));
}